import sys
from math import floor, ceil


class Converge():
	'''
	Utility to find valid lower & upper bounds of list.
	'''

	def __init__(self, start, end, callback):
		'''
		Pass random start and end point (both must be vlid, i.e, must be inside bounds),
		and a difference greater than 1 i.e. abs(end-start) > 1,
		with a callback method that will verify whether a prediction is out of bounds or not.
		:param start: lower (known) random bound
		:param end: higher (known) random bound
		:param callback: Verification method
		'''
		self.low = start
		self.high = end
		self.lowStep = (start - end) / 2
		self.highStep = floor((end - start) / 2)
		self.verify = callback

	def run(self):
		while True:  # Converge on low
			IsLowInBounds = self.verify(self.low + self.lowStep)
			if IsLowInBounds:
				self.low = self.low + self.lowStep
				self.lowStep *= 2  # Move faster
			else:
				half = self.lowStep / (2 * 1.0)
				if half < 0:
					half = ceil(half)
				else:
					half = floor(half)
				self.lowStep = half  # Move slower

			if self.lowStep == 0:
				break

		while True:  # Converge on high
			IsHighInBounds = self.verify(self.high + self.highStep)
			if IsHighInBounds:
				self.high = self.high + self.highStep
				self.highStep *= 2  # Move faster
			else:
				half = self.highStep / (2 * 1.0)
				if half < 0:
					half = ceil(half)
				else:
					half = floor(half)
				self.highStep = half  # Move slower

			if self.highStep == 0:
				break

		# IsHighInBounds = self.verify(self.high)
		return (self.low, self.high)


if __name__ == "__main__":
	def check_validity(val):
		if val > -30 and val < 101:
			return True
		else:
			return False


	low, high = Converge(-28, 50, check_validity).run()
	pass



import win32com.client
import pythoncom


class Handler_Class(object):
	def OnNewMailEx(self, receivedItemsIDs):
		for ID in receivedItemsIDs.split(","):
			# Microsoft.Office.Interop.Outlook _MailItem properties:
			# https://msdn.microsoft.com/en-us/library/microsoft.office.interop.outlook._mailitem_properties.aspx
			mailItem = MailBot.outlookHandled.Session.GetItemFromID(ID)
			print "Subj: " + mailItem.Subject
			print "Body: " + mailItem.Body.encode('ascii', 'ignore')
			print "========"


class MailBot():
	outlookHandled = win32com.client.DispatchWithEvents("Outlook.Application", Handler_Class)

	def __init__(self):
		self.outlook = win32com.client.Dispatch(
			"Outlook.Application")  # https://msdn.microsoft.com/en-us/vba/outlook-vba/articles/application-object-outlook#
		self.application = self.outlook.Application
		self.namespace = self.application.GetNamespace("MAPI")
		self.inbox = self.namespace.GetDefaultFolder(6)

	def eval(self, object, param):
		res = None
		try:
			res = eval("object.%s" % (param))
		except AttributeError as e:
			pass

		return res

	def BuildMailItem(self, message):
		resItem = dict()
		resItem["message"] = message
		resItem["Subject"] = self.eval(message, "Subject")
		resItem["Body"] = self.eval(message, "Body")
		resItem["HTMLBody"] = self.eval(message, "HTMLBody")
		resItem["Importance"] = self.eval(message, "Importance")
		resItem["Recipients"] = self.eval(message, "Recipients")
		resItem["Sender"] = self.eval(message, "Sender")
		resItem["To"] = self.eval(message, "To")
		resItem["CC"] = self.eval(message, "CC")
		resItem["BCC"] = self.eval(message, "BCC")
		resItem["SenderEmailAddress"] = self.eval(message, "SenderEmailAddress")
		resItem["SenderName"] = self.eval(message, "SenderName")
		resItem["SentOn"] = self.eval(message, "SentOn")
		resItem["Size"] = self.eval(message, "Size")
		resItem["UnRead"] = self.eval(message, "UnRead")
		resItem["VotingOption"] = self.eval(message, "VotingOption")
		resItem["VotingResponse"] = self.eval(message, "VotingResponse")
		resItem["Attachments"] = self.eval(message, "Attachments")
		return resItem

	def GetAllMails(self):
		"""
		# Reference: https://stackoverflow.com/questions/22813814/clearly-documented-reading-of-emails-functionality-with-python-win32com-outlook
		:return: all the mails in the inbox
		"""
		messages = self.inbox.Items

		res = []
		for message in messages:
			resItem = self.BuildMailItem(message)
			res.append(resItem)
		return res

	def GetLastMail(self):
		message = self.inbox.Items.GetLast()
		res = self.BuildMailItem(message)
		return res

	def GetAllFolders(self):
		for i in range(50):
			try:
				box = self.namespace.GetDefaultFolder(i)
				name = box.Name
				print(i, name)
			except:
				pass

	def Search(self, query):
		strF = "urn:schemas:mailheader:subject = '%s'" % (query)
		strS = "Inbox"
		results = self.application.AdvancedSearch(strS, strF).Results
		for result in [results.Item(i) for i in range(1, results.Count)]:
			print(result)
		return results

	def SendMail(self, mail=None, To=None, CC=None, BCC=None, Subject=None,
				 Body=None, Attachments=None,
				 draft=True):
		"""
		Reference: https://stackoverflow.com/questions/17449184/sending-outlook-email-with-body-as-contents-of-a-text-file
		:param mail: Mail object to reply to.
		:param BCC:
		:param CC:
		:param Attachments: List of attachments
		:param To:
		:param Subject:
		:param Body:
		:param draft: Set it to false if the mail is to be sent automatically.
		:return:
		"""
		if mail is None:
			mail = self.application.CreateItem(0)

		if To is None:
			To = ["user@user.com", "user2@user.com"]
		if Attachments is None:
			Attachments = []

		if To:
			for toItem in To:
				recipient = mail.Recipients.Add(toItem)
				recipient.Type = 1
		if CC:
			for ccItem in CC:
				recipient = mail.Recipients.Add(ccItem)
				recipient.Type = 2
		if BCC:
			for bccItem in BCC:
				recipient = mail.Recipients.Add(bccItem)
				recipient.Type = 3
		if Subject:
			mail.Subject = Subject
		if Body:
			mail.Body = Body
		if Attachments:
			for attachment in Attachments:
				mail.Attachments.Add(attachment)
		if draft:
			mail.Display()
		else:
			mail.Send()

	def ReplyToMail(self, mail, replyToAll=False, To=None, CC=None, BCC=None,
					Body=None, Attachments=None,
					draft=True):
		"""
		:param mail: Mail object to reply to.
		:param BCC:
		:param CC:
		:param Attachments: List of attachments
		:param To:
		:param Body:
		:param draft: Set it to false if the mail is to be sent automatically.
		:return:
		"""
		if replyToAll:
			newMail = mail.ReplyAll()
		else:
			newMail = mail.Reply()
		Body = Body + "\n" + newMail.Body
		self.SendMail(newMail, To=To, CC=CC, BCC=BCC, Body=Body, Attachments=Attachments, draft=draft)

	def ForwardMail(self, mail, To=None, CC=None, BCC=None,
					Body=None, Attachments=None,
					draft=True):
		"""
		:param mail: Mail object to be forwarded
		:param BCC:
		:param CC:
		:param Attachments: List of attachments
		:param To:
		:param Body:
		:param draft: Set it to false if the mail is to be sent automatically.
		:return:
		"""
		newMail = mail.Forward()
		Body = Body + "\n" + newMail.Body
		self.SendMail(newMail, To=To, CC=CC, BCC=BCC, Body=Body, Attachments=Attachments, draft=draft)

	def SwitchUnread(self, mail):
		if mail.UnRead == True:
			mail.UnRead = False
		else:
			mail.UnRead = True


if __name__ == "__main__":
	last = MailBot().GetLastMail()
	# MailBot().SendMail(
	# 	Subject="Test mail",
	# 	Body="This is the test body.\n \ncan go multiline",
	# 	To=["xyz@abc.com", "xyz@abc.com"],
	# 	CC=["xyz@abc.com", "xyz@abc.com"],
	# 	BCC=["xyz@abc.com", "xyz@abc.com"],
	# 	Attachments=["D:\xyz\abc\foo.txt"]
	# )

	MailBot().ReplyToMail(
		mail=last["message"],
		# replyToAll=True,
		Body="This is the test body.\n \ncan go multiline",
		To=["xyz@abc.com", "xyz@abc.com"],
		CC=["xyz@abc.com", "xyz@abc.com"],
		BCC=["xyz@abc.com", "xyz@abc.com"],
		Attachments=["D:\xyz\abc\foo.txt"]
	)

	# MailBot().ForwardMail(
	# 	mail=last["message"],
	# 	Body="This is the test body.\n \ncan go multiline",
	# 	To=["xyz@abc.com", "xyz@abc.com"],
	# 	CC=["xyz@abc.com", "xyz@abc.com"],
	# 	BCC=["xyz@abc.com", "xyz@abc.com"],
	# 	Attachments=["D:\xyz\abc\foo.py"]
	# )

	# search = MailBot().Search("Test")
	# folders = MailBot().GetAllFolders()
	# mails = MailBot().GetAllMails()
	pass



import os
import sys

from app import app, db
from flask_cors import CORS, cross_origin
from app.data.config import Config
from werkzeug.contrib.fixers import ProxyFix

from scripts.create_db_schema import initDB

config = Config()

'''Run app'''
CORS(app, origins=config["allowed_domains"])  # Allow cross-domain
app.wsgi_app = ProxyFix(app.wsgi_app)

initDB()

if __name__ == '__main__':
    debug = config["debug"]
    host = config["server"]["host"]
    port = config["server"]["port"]
    app.run(port=port, debug=debug, host=host, threaded=True)



# Install and config elasticsearch :
# 	http://caraulean.com/2016/install-and-configure-elasticsearch-in-windows/
# 	https://www.elastic.co/downloads/elasticsearch

# !/usr/bin/python3
#   This script is a simple introduction to the python elasticsearch API.
#
#   This script will populate an elasticsearch index from a file and then give a simple command line query interface.
#   Each line of the input file will be mapped into a JSON document of the form { "text": "my file line..." } and added
#   to the index.
#
#   You can use Docker to spin up a local elasticsearch instance to play around with, e.g.
#   docker run --name elasticsearch -d -p 9200:9200 elasticsearch:latest
#

import argparse, elasticsearch, json

import os
from elasticsearch import Elasticsearch

from ir_data import IRData


def parse_args():
	# argument help
	parser = argparse.ArgumentParser(description='Add lines from a file to a simple text Elasticsearch index.')
	parser.add_argument('--file', help='Name of file to parse, e.g. /usr/share/dict/american-english')
	parser.add_argument('--host', default="localhost", help='Elasticsearch host.')
	parser.add_argument('-p', '--port', default=9200, help='port, default is 9200')
	return parser.parse_args()


def index_irs(NBR_OF_IRS=1000):
	latest = IRData().GetLatestIR()
	for incident in IRData().GetListOfIRs(latest - NBR_OF_IRS, latest):
		print("Indexing - %s" % (incident))
		es.index(
			index="irs",
			doc_type="document",
			body=incident,
			# parent=currentParent["id"]
		)


def search():
	# now we can do searches.
	print("Ok. I've got an index of {0} documents. Let's do some searches...".format(count['count']))
	while True:
		try:
			query = input("Enter a search: ")
			result = es.search(index=INDEX_NAME, doc_type=TYPE,
							   body=fuzzy_search(query, "eco_justification_free_text"))
			if result.get('hits') is not None and result['hits'].get('hits') is not None:
				print("\nResults: " + str(result['hits']['total']) + "\n")
				for hit in result['hits']['hits']:
					source = hit["_source"]
					print("name: " + str(source["name"][0]))
					print("title: " + str(source["title"][0]))
					# print("description: " + source["description"][0].encode('utf-8'))
					# print("severity: " + str(source["severity"][0]))
					print("lastmodifieddate: " + str(source["lastmodifieddate"][0]) + "\n")
			else:
				print({})
		except(KeyboardInterrupt):
			break
		except KeyError as e:
			pass


def fuzzy_search(term, field):
	return {
		"query": {
			"match": {
				field: {
					"query": term.strip(),
					"fuzziness": "AUTO"
				}
			}
		}
	}


if __name__ == "__main__":
	args = parse_args()

	# index and document type constants
	INDEX_NAME = "irs"
	TYPE = "document"

	# get a client
	es = Elasticsearch(hosts=[{"host": args.host, "port": args.port}])

	# create an index, ignore if it exists already
	es.indices.create(index='documents', ignore=400)

	# index_irs(10000)

	# count the matches
	count = es.count(index=INDEX_NAME, doc_type=TYPE, body={"query": {"match_all": {}}})

	search()



import json
import os
import uuid

from dukepy.dict_diff import dict_diff
from dukepy.singleton import Singleton


class ConfigError(Exception):
    pass


class Config(dict, metaclass=Singleton):
    def __str__(self):
        return json.dumps(self.config, indent=4)

    def __init__(self, path="C:\\temp\\config.json", defaults=None):
        """
        :param path: Full path of the json file
        :param defaults: Default configuration eg. {"key1":"value1", "keygroup1":{"key2":"value2"}}
        """

        dict.__init__(self, dict())  # Because dict is extended

        if defaults is None:
            self.defaults = self._defaults()
        else:
            self.defaults = defaults

        self.config_file = os.path.normpath(path)
        self.config_dir = os.path.dirname(self.config_file)

        # create config.json and initialize empty self.config
        if not os.path.exists(self.config_file):
            if not os.path.exists(self.config_dir):
                os.mkdir(self.config_dir)  # Create directory
            open(self.config_file, "a").close()  # Create empty file
            self.config = dict()
            self.config = self.defaults  # Initialize with default values
            self.commit()

        # initialize self.config from config.json
        with open(self.config_file, "r+") as f:
            conf = f.read()
            if conf != "":
                try:
                    config_stored = json.loads(conf)
                    config_deafult = self.defaults
                    are_config_keys_modified = dict_diff(config_deafult, config_stored,
                                                         udpate_modified_keys=True,
                                                         udpate_added_keys=False,
                                                         udpate_removed_keys=False)

                    self.config = config_deafult
                    if are_config_keys_modified:
                        self.commit()
                except json.decoder.JSONDecodeError as e:
                    raise ConfigError("Config file invalid format")
            else:
                self.config = dict()

    def __enter__(self):
        """ to enable 'with **' capability, counterpart function is __exit__ """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ commit at the end of 'with **' block """
        self.commit()

    def __getitem__(self, key):
        # return super().__getitem__(key)
        try:
            return self.config[key]
        except KeyError as e:
            self.__setitem__(key, dict())  # initialize the non-existent keys
            return self.config[key]

    def __setitem__(self, key, value):
        # super().__setitem__(key, value)
        self.config[key] = value

    def commit(self):
        """
        Commit the configuration changes to file
        Use "with Config() as config" if auto commit is needed at the end,
        otherwise use this method.
        """
        # make a copy of the original config
        from shutil import copyfile
        copyfile(self.config_file, self.config_file + "." + str(uuid.uuid4()))

        # overwrite the file with new config
        with open(self.config_file, "w+") as f:
            json.dump(self.config, f, indent=4)

    def _defaults(self):
        config_default = dict()

        sqlite_path = os.path.join(self.config_dir, "sqlite.db")

        config_default["server"] = {
            "host": "0.0.0.0",
            "port": "80"
        }
        config_default["database"] = {
            "active": "sqlite",
            "mysql": {
                "db": "dummy_db",
                "user": "dummy_user",
                "host": "localhost",
                "password": "dummy_password"
            },
            "postgres": {
                "db": "dummy_db",
                "user": "dummy_user",
                "host": "localhost",
                "password": "dummy_password",
                "port": "5432"
            },
            "sqlite": {
                "path": sqlite_path
            },
            "firebase": {
                "service_account_key": "path_to_serviceAccountKey.json",
                "databaseURL": "https://xyz_project_123.firebaseio.com"
            }
        }
        config_default["debug"] = True
        config_default["stacktrace"] = True
        config_default["allowed_domains"] = ["http://1", "http://2"]

        return config_default


if __name__ == "__main__":
    with Config() as config:
        pass

        # config.commit() #Not required, since we are using with*



# HTTP Error Codes and their meaning: https://www.symantec.com/connect/articles/http-error-codes

from flask import Flask
from flask_sqlalchemy import SQLAlchemy

from app.data.config import Config
from app.data.tables.alchemy_base import AlchemyBase, db_uri, db_session

config = Config()

""" App """
## Init app
app = Flask(__name__)

## Init SQLAlchemy
app.config['SQLALCHEMY_DATABASE_URI'] = db_uri
if config["debug"]:
    app.config['SQLALCHEMY_ECHO'] = True
db = SQLAlchemy(app, model_class=AlchemyBase)
# db.create_all() # No use until the model classes are imported. Use scripts.create_db_schema instead.

""" Views """
from app.views import ...



import datetime
import traceback
from enum import Enum

from sqlalchemy import create_engine, Column, DateTime, func, Integer, BigInteger, String
from sqlalchemy.dialects.postgresql import JSON, JSONB
from sqlalchemy.exc import DatabaseError
from flask_sqlalchemy import Model
from sqlalchemy.orm import sessionmaker

from app import Config
from app.utils.traces import print_exception_traces


class DBType(Enum):
    sqlite = 1
    mysql = 2
    postgres = 3


db_type = DBType.sqlite  # Default
if "sqlite" == Config()["database"]["active"]:
    db_type = DBType.sqlite
if "mysql" == Config()["database"]["active"]:
    db_type = DBType.mysql
if "postgres" == Config()["database"]["active"]:
    db_type = DBType.postgres


## Init session
class DBSession():
    def uri(self, type):
        db_uri = None

        if type == DBType.sqlite:
            db_uri = "sqlite:///" + Config()["database"]["sqlite"]["path"]

        if type == DBType.mysql:
            config = Config()["database"]["mysql"]
            db_uri = "mysql+pymysql://{0}:{1}@{2}:3306/{3}".format(config["user"], config["password"], config["host"],
                                                                   config["db"])

        if type == DBType.postgres:
            config = Config()["database"]["postgres"]
            db_uri = "postgresql://{0}:{1}@{2}:{3}/{4}".format(config["user"], config["password"], config["host"],
                                                               config["port"], config["db"])

        return db_uri

    def create_session(self, db_uri):
        engine = create_engine(db_uri)
        Session = sessionmaker(bind=engine)
        session = Session()
        session._model_changes = {}
        return session


db_session_helper = DBSession()
db_uri = db_session_helper.uri(db_type)
db_session = db_session_helper.create_session(db_uri)


## Declare base
class AlchemyBase(Model):
    """
    ref: https://chase-seibert.github.io/blog/2016/03/31/flask-sqlalchemy-sessionless.html
    """

    if db_type == DBType.sqlite:
        id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
        json_data = Column(String, nullable=True)
    if db_type == DBType.mysql:
        id = Column(BigInteger, primary_key=True, autoincrement=True, nullable=False)
        json_data = Column(String, nullable=True)
    if db_type == DBType.postgres:
        id = Column(BigInteger, primary_key=True, autoincrement=True, nullable=False)
        json_data = Column(JSONB, nullable=True)

    # https://stackoverflow.com/a/12155686/973425
    created_at = Column(DateTime, nullable=False,
                        server_default=func.now())
    updated_at = Column(DateTime, nullable=False,
                        server_default=func.now(),
                        server_onupdate=func.now())

    def save(self):
        local_object = db_session.merge(self)
        db_session.add(local_object)
        # try:
        self._flush()
        db_session.commit()
        # except DatabaseError as e:
        #     code = e.orig.args[0]
        #     if code == 1062:
        #         raise
        #     return None
        return local_object

    def update(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        return self.save()

    def delete(self):
        local_object = db_session.merge(self)  # https://stackoverflow.com/a/47663833/973425
        db_session.delete(local_object)
        self._flush()
        db_session.commit()

    def _flush(self):
        try:
            db_session.flush()
            # db_session.refresh(self)
        except DatabaseError as e:
            db_session.rollback()
            print_exception_traces(e)
            raise

    @staticmethod
    def session():
        return db_session



import os

import setuptools

file_path = os.path.abspath(__file__)
readme = os.path.join(os.path.dirname(file_path), "README.md")
# print(readme)
with open(readme, "r") as fh:
    long_description = fh.read()

setuptools.setup(name='dukepy',
                 description='A useful collection of utilities',
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 version='0.0.6',
                 url='https://github.com/duke79/libpython',
                 author='Pulkit Singh',
                 author_email='pulkitsingh01@gmail.com',
                 license='MIT License',
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 packages=setuptools.find_packages(),
                 install_requires=[
                     'colorama',
                     'enum'
                 ],
                 entry_points={
                     'console_scripts': [
                         'encrypt=dukepy.main:run'
                     ]
                 }
                 )



from colorama import Fore, Style
from enum import Enum
from functools import total_ordering


class Logger:
	indents = 0
	indentsFlag = True

	@total_ordering
	class Levels:
		# __order__ = "CRITICAL FATAL ERROR SUCCESS WARNING NOTICE INFO VERBOSE DEBUG SPAM ALL"
		NONE = 100
		""" No log is printed if level==NONE"""
		OUTPUT = 55
		CRITICAL = 50
		FATAL = CRITICAL
		ERROR = 40
		SUCCESS = 35
		WARNING = 30
		NOTICE = 25
		INFO = 20
		VERBOSE = 15
		DEBUG = 10
		SPAM = 5
		ALL = 0

		def __lt__(self, other):
			if self.__class__ is other.__class__:
				return self.value < other.value

	level = Levels.ALL
	"""The logs levels below this level are filtered. eg. if level==Level.DEBUG then spam() won't print anything."""

	def __init__(self):
		import colorama
		colorama.init()
		pass

	def spam(self, msg):
		if self.level <= self.Levels.SPAM:
			print(Fore.WHITE + "SPAM: " + msg + Style.RESET_ALL)

	def debug(self, msg, indentOffset=0):
		"""
		Print debug logs.
		:param msg: The string to print.
		:param indentOffset: Increment/Decrement in indent space (must provide at least 0, if indent to be printed)
		:return:
		"""
		if self.level <= self.Levels.DEBUG:
			if indentOffset:
				self.printIndent(indentOffset)
			print(Fore.GREEN + "DEBUG: " + msg + Style.RESET_ALL)

	def verbose(self, msg):
		if self.level <= self.Levels.VERBOSE:
			print(Fore.MAGENTA + "VERBOSE: " + msg + Style.RESET_ALL)

	def info(self, msg):
		if self.level <= self.Levels.INFO:
			print(Fore.WHITE + Style.BRIGHT + "INFO: " + msg + Style.RESET_ALL)

	def notice(self, msg):
		if self.level <= self.Levels.NOTICE:
			print(Fore.RED + "NOTICE: " + msg + Style.RESET_ALL)

	def warning(self, msg):
		if self.level <= self.Levels.WARNING:
			print(Fore.YELLOW + Style.BRIGHT + "WARNING: " + msg + Style.RESET_ALL)

	def success(self, msg):
		if self.level <= self.Levels.SUCCESS:
			print(Fore.GREEN + Style.BRIGHT + "SUCCESS: " + msg + Style.RESET_ALL)

	def error(self, msg):
		if self.level <= self.Levels.ERROR:
			print(Fore.RED + Style.BRIGHT + "ERROR: " + msg + Style.RESET_ALL)

	def fatal(self, msg):
		if self.level <= self.Levels.FATAL:
			print(Fore.MAGENTA + Style.BRIGHT + "FATAL: " + msg + Style.RESET_ALL)

	def critical(self, msg):
		if self.level <= self.Levels.CRITICAL:
			print(Fore.MAGENTA + Style.BRIGHT + "CRITICAL: " + msg + Style.RESET_ALL)

	def output(self, msg):
		if self.level <= self.Levels.OUTPUT:
			try:
				print(msg)
			except UnicodeEncodeError as e:
				pass
			except IOError as e:
				pass

	def printIndent(self, indentOffset):
		if indentOffset > 0:
			self.indents += indentOffset

		if self.indents < 0:
			self.indents = 0
		indentSpace = ""
		for i in range(0, self.indents):
			indentSpace += "  "
		print(indentSpace)

		if indentOffset < 0:
			self.indents += indentOffset



from sqlalchemy.orm import relationship
from sqlalchemy.sql import expression

from app import db
from sqlalchemy.dialects.mysql import INTEGER


class User(db.Model):
    __tablename__ = 'user'    
    name = db.Column(db.String(200), nullable=True)
    phone_number = db.Column(db.String(50), nullable=True)
    photo_url = db.Column(db.String(2013), nullable=True)
    email = db.Column(db.String(200), nullable=True)
    firebase_uid = db.Column(db.String(100), nullable=True)
    admin = db.Column(db.Boolean, nullable=False, server_default=expression.false())    
    # children = relationship("user_categories")

    def __repr__(self):
        return '<User %r>' % self.username



def dict_diff(d1, d2,
              udpate_modified_keys=False,
              udpate_added_keys=False,
              udpate_removed_keys=False,
              path=""):
    """

    :param d1:
    :param d2:
    :param udpate_modified_keys:
    :param udpate_added_keys:
    :param udpate_removed_keys:
    :param path:
    :return: True if at least one difference.
    """
    ret = False
    for k in list(d1.keys()):
        if not k in d2:
            ret = True
            print(path, ":")
            # print(k + " as key not in d2", "\n")
            print(" - ", k, " : ", d1[k])
            if udpate_removed_keys:
                d1.pop(k)
        else:
            if type(d1[k]) is dict:
                if path == "":
                    local_path = k
                else:
                    local_path = path + "->" + k

                ret = dict_diff(d1[k], d2[k],
                                udpate_modified_keys=udpate_modified_keys,
                                udpate_added_keys=udpate_added_keys,
                                udpate_removed_keys=udpate_removed_keys,
                                path=local_path)
            else:
                if d1[k] != d2[k]:
                    # ret = True
                    print(path, ":")
                    print(" - ", k, " : ", d1[k])
                    print(" + ", k, " : ", d2[k])
                    if udpate_modified_keys:
                        d1[k] = d2[k]
    for k in list(d2.keys()):
        if not k in d1:
            ret = True
            print(path, ":")
            print(" + ", k, " : ", d2[k])
            if udpate_added_keys:
                d1[k] = d2[k]

    return ret
